@@ -310,6 +310,10 @@ class DatabaseWrapper(BaseDatabaseWrapper):
         else:
             return True
 
+    @cached_property
+    def cx_oracle_version(self):
+        return tuple(int(x) for x in Database.version.split('.'))
+
     @cached_property
     def oracle_version(self):
         with self.temporary_connection():
